package com.sun.java.util.jar.pack;

import java.io.OutputStream;
import java.util.HashMap;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

class Coding implements Comparable<Coding>, CodingMethod, Histogram.BitMetric
{
    public static final int B_MAX = 5;
    public static final int H_MAX = 256;
    public static final int S_MAX = 2;
    private final int B;
    private final int H;
    private final int L;
    private final int S;
    private final int del;
    private final int min;
    private final int max;
    private final int umin;
    private final int umax;
    private final int[] byteMin;
    private final int[] byteMax;
    private static Map<Coding, Coding> codeMap;
    private static final byte[] byteBitWidths;
    static boolean verboseStringForDebug;
    
    private static int saturate32(long n) {
        if (n > 2147483647L) {
            return Integer.MAX_VALUE;
        }
        if (n < -2147483648L) {
            return Integer.MIN_VALUE;
        }
        return (int)n;
    }
    
    private static long codeRangeLong(int n, int n2) {
        return codeRangeLong(n, n2, n);
    }
    
    private static long codeRangeLong(int n, int n2, int n3) {
        assert n3 >= 0 && n3 <= n;
        assert n >= 1 && n <= 5;
        assert n2 >= 1 && n2 <= 256;
        if (n3 == 0) {
            return 0L;
        }
        if (n == 1) {
            return (long)n2;
        }
        int n4 = 256 - n2;
        long n5 = 0L;
        long n6 = 1L;
        for (int i = 1; i <= n3; ++i) {
            n5 += n6;
            n6 *= (long)n2;
        }
        long n7 = n5 * (long)n4;
        if (n3 == n) {
            n7 += n6;
        }
        return n7;
    }
    
    public static int codeMax(int n, int n2, int n3, int n4) {
        long codeRangeLong = codeRangeLong(n, n2, n4);
        if (codeRangeLong == 0L) {
            return -1;
        }
        if (n3 == 0 || codeRangeLong >= 4294967296L) {
            return saturate32(codeRangeLong - 1L);
        }
        long n5;
        for (n5 = codeRangeLong - 1L; isNegativeCode(n5, n3); --n5) {}
        if (n5 < 0L) {
            return -1;
        }
        int decodeSign32 = decodeSign32(n5, n3);
        if (decodeSign32 < 0) {
            return Integer.MAX_VALUE;
        }
        return decodeSign32;
    }
    
    public static int codeMin(int n, int n2, int n3, int n4) {
        long codeRangeLong = codeRangeLong(n, n2, n4);
        if (codeRangeLong >= 4294967296L && n4 == n) {
            return Integer.MIN_VALUE;
        }
        if (n3 == 0) {
            return 0;
        }
        long n5;
        for (n5 = codeRangeLong - 1L; !isNegativeCode(n5, n3); --n5) {}
        if (n5 < 0L) {
            return 0;
        }
        return decodeSign32(n5, n3);
    }
    
    private static long toUnsigned32(int n) {
        return (long)n << 32 >>> 32;
    }
    
    private static boolean isNegativeCode(long n, int n2) {
        assert n2 > 0;
        assert n >= -1L;
        return ((int)n + 1 & (1 << n2) - 1) == 0x0;
    }
    
    private static boolean hasNegativeCode(int n, int n2) {
        assert n2 > 0;
        return 0 > n && n >= ~(-1 >>> n2);
    }
    
    private static int decodeSign32(long i, int n) {
        assert i == toUnsigned32((int)i) : Long.toHexString(i);
        if (n == 0) {
            return (int)i;
        }
        int n2;
        if (isNegativeCode(i, n)) {
            n2 = ~((int)i >>> n);
        }
        else {
            n2 = (int)i - ((int)i >>> n);
        }
        assert n2 == ((int)i >>> 1 ^ -((int)i & 0x1));
        return n2;
    }
    
    private static long encodeSign32(int i, int n) {
        if (n == 0) {
            return toUnsigned32(i);
        }
        int n2 = (1 << n) - 1;
        long n3;
        if (!hasNegativeCode(i, n)) {
            n3 = (long)i + toUnsigned32(i) / (long)n2;
        }
        else {
            n3 = (long)((-i << n) - 1);
        }
        long unsigned32 = toUnsigned32((int)n3);
        assert i == decodeSign32(unsigned32, n) : Long.toHexString(unsigned32) + " -> " + Integer.toHexString(i) + " != " + Integer.toHexString(decodeSign32(unsigned32, n));
        return unsigned32;
    }
    
    public static void writeInt(byte[] array, int[] array2, int n, int n2, int n3, int n4) {
        long encodeSign32 = encodeSign32(n, n4);
        assert encodeSign32 == toUnsigned32((int)encodeSign32);
        assert encodeSign32 < codeRangeLong(n2, n3) : Long.toHexString(encodeSign32);
        int n5 = 256 - n3;
        int n6 = array2[0];
        long n8;
        int n9;
        for (int n7 = 0; n7 < n2 - 1 && encodeSign32 >= (long)n5; encodeSign32 = n8 / (long)n3, array[n6++] = (byte)n9, ++n7) {
            n8 = encodeSign32 - (long)n5;
            n9 = (int)((long)n5 + n8 % (long)n3);
        }
        array[n6++] = (byte)(int)encodeSign32;
        array2[0] = n6;
        return;
    }
    
    public static int readInt(byte[] array, int[] array2, int n, int n2, int n3) {
        int n4 = 256 - n2;
        long n5 = 0L;
        long n6 = 1L;
        int n7 = array2[0];
        int i = 0;
        while (i < n) {
            int n8 = array[n7++] & 0xFF;
            n5 += (long)n8 * n6;
            n6 *= (long)n2;
            if (n8 < n4) {
                break;
            }
            ++i;
        }
        array2[0] = n7;
        return decodeSign32(n5, n3);
    }
    
    public static int readIntFrom(InputStream inputStream, int n, int n2, int n3) throws IOException {
        int n4 = 256 - n2;
        long n5 = 0L;
        long n6 = 1L;
        int i = 0;
        while (i < n) {
            int read = inputStream.read();
            if (read < 0) {
                throw new RuntimeException("unexpected EOF");
            }
            n5 += (long)read * n6;
            n6 *= (long)n2;
            if (read < n4) {
                break;
            }
            ++i;
        }
        assert n5 >= 0L && n5 < codeRangeLong(n, n2);
        return decodeSign32(n5, n3);
    }
    
    private Coding(int n, int n2, int n3) {
        this(n, n2, n3, 0);
    }
    
    private Coding(int b, int h, int s, int del) {
        this.B = b;
        this.H = h;
        this.L = 256 - h;
        this.S = s;
        this.del = del;
        this.min = codeMin(b, h, s, b);
        this.max = codeMax(b, h, s, b);
        this.umin = codeMin(b, h, 0, b);
        this.umax = codeMax(b, h, 0, b);
        this.byteMin = new int[b];
        this.byteMax = new int[b];
        for (int i = 1; i <= b; ++i) {
            this.byteMin[i - 1] = codeMin(b, h, s, i);
            this.byteMax[i - 1] = codeMax(b, h, s, i);
        }
    }
    
    @Override
    public boolean equals(Object o) {
        if (!(o instanceof Coding)) {
            return false;
        }
        Coding coding = (Coding)o;
        return this.B == coding.B && this.H == coding.H && this.S == coding.S && this.del == coding.del;
    }
    
    @Override
    public int hashCode() {
        return (this.del << 14) + (this.S << 11) + (this.B << 8) + (this.H << 0);
    }
    
    private static synchronized Coding of(int n, int n2, int n3, int n4) {
        if (Coding.codeMap == null) {
            Coding.codeMap = new HashMap();
        }
        Coding coding = new Coding(n, n2, n3, n4);
        Coding coding2 = (Coding)Coding.codeMap.get(coding);
        if (coding == null) {
            Coding.codeMap.put(coding, coding);
            return coding;
        }
        return coding;
    }
    
    public static Coding of(int n, int n2) {
        return of(n, n2, 0, 0);
    }
    
    public static Coding of(int n, int n2, int n3) {
        return of(n, n2, n3, 0);
    }
    
    public boolean canRepresentValue(int n) {
        if (this.isSubrange()) {
            return this.canRepresentUnsigned(n);
        }
        return this.canRepresentSigned(n);
    }
    
    public boolean canRepresentSigned(int n) {
        return n >= this.min && n <= this.max;
    }
    
    public boolean canRepresentUnsigned(int n) {
        return n >= this.umin && n <= this.umax;
    }
    
    public int readFrom(byte[] array, int[] array2) {
        return readInt(array, array2, this.B, this.H, this.S);
    }
    
    public void writeTo(byte[] array, int[] array2, int n) {
        writeInt(array, array2, n, this.B, this.H, this.S);
    }
    
    public int readFrom(InputStream inputStream) throws IOException {
        return readIntFrom(inputStream, this.B, this.H, this.S);
    }
    
    public void writeTo(OutputStream outputStream, int n) throws IOException {
        byte[] b = new byte[this.B];
        int[] array = { 0 };
        writeInt(b, array, n, this.B, this.H, this.S);
        outputStream.write(b, 0, array[0]);
    }
    
    @Override
    public void readArrayFrom(InputStream inputStream, int[] array, int i, int n) throws IOException {
        for (; i < n; ++i) {
            array[i] = this.readFrom(inputStream);
        }
        for (int j = 0; j < this.del; ++j) {
            long n2;
            for (n2 = 0L, ; i < n; array[i] = (int)n2, ++i) {
                n2 += (long)array[i];
                if (this.isSubrange()) {
                    n2 = (long)this.reduceToUnsignedRange(n2);
                }
            }
        }
    }
    
    @Override
    public void writeArrayTo(OutputStream outputStream, int[] array, int i, int length) throws IOException {
        if (length <= i) {
            return;
        }
        int[] array2;
        for (int j = 0; j < this.del; array = array2, i = 0, length = array2.length, ++j) {
            if (!this.isSubrange()) {
                array2 = makeDeltas(array, i, length, 0, 0);
            }
            else {
                array2 = makeDeltas(array, i, length, this.min, this.max);
            }
        }
        byte[] b = new byte[256];
        int n = b.length - this.B;
        int[] array3;
        for (array3 = new int[] { 0 }, ; i < length; array3[0] = 0) {
            while (array3[0] <= n) {
                this.writeTo(b, array3, array[i++]);
                if (i >= length) {
                    break;
                }
                continue;
            }
            outputStream.write(b, 0, array3[0]);
        }
        return;
    }
    
    boolean isSubrange() {
        return this.max < Integer.MAX_VALUE && (long)this.max - (long)this.min + 1L <= 2147483647L;
    }
    
    boolean isFullRange() {
        return this.max == Integer.MAX_VALUE && this.min == Integer.MIN_VALUE;
    }
    
    int getRange() {
        assert !(!this.isSubrange());
        return this.max - this.min + 1;
    }
    
    Coding setB(int n) {
        return of(n, this.H, this.S, this.del);
    }
    
    Coding setH(int n) {
        return of(this.B, n, this.S, this.del);
    }
    
    Coding setS(int n) {
        return of(this.B, this.H, n, this.del);
    }
    
    Coding setL(int n) {
        return this.setH(256 - n);
    }
    
    Coding setD(int n) {
        return of(this.B, this.H, this.S, n);
    }
    
    Coding getDeltaCoding() {
        return this.setD(this.del + 1);
    }
    
    Coding getValueCoding() {
        if (this.isDelta()) {
            return of(this.B, this.H, 0, this.del - 1);
        }
        return this;
    }
    
    int reduceToUnsignedRange(long n) {
        if (n == (long)(int)n && this.canRepresentUnsigned((int)n)) {
            return (int)n;
        }
        int range = this.getRange();
        assert range > 0;
        n %= (long)range;
        if (n < 0L) {
            n += (long)range;
        }
        assert !(!this.canRepresentUnsigned((int)n));
        return (int)n;
    }
    
    int reduceToSignedRange(int n) {
        if (this.canRepresentSigned(n)) {
            return n;
        }
        return reduceToSignedRange(n, this.min, this.max);
    }
    
    static int reduceToSignedRange(int n, int n2, int n3) {
        int n4 = n3 - n2 + 1;
        assert n4 > 0;
        n -= n2;
        if (n < 0 && n >= 0) {
            n -= n4;
            assert n >= 0;
        }
        n %= n4;
        if (n < 0) {
            n += n4;
        }
        n += n2;
        assert n2 <= n && n <= n3;
        return n;
    }
    
    boolean isSigned() {
        return this.min < 0;
    }
    
    boolean isDelta() {
        return this.del != 0;
    }
    
    public int B() {
        return this.B;
    }
    
    public int H() {
        return this.H;
    }
    
    public int L() {
        return this.L;
    }
    
    public int S() {
        return this.S;
    }
    
    public int del() {
        return this.del;
    }
    
    public int min() {
        return this.min;
    }
    
    public int max() {
        return this.max;
    }
    
    public int umin() {
        return this.umin;
    }
    
    public int umax() {
        return this.umax;
    }
    
    public int byteMin(int n) {
        return this.byteMin[n - 1];
    }
    
    public int byteMax(int n) {
        return this.byteMax[n - 1];
    }
    
    @Override
    public int compareTo(Coding coding) {
        int n = this.del - coding.del;
        if (n == 0) {
            n = this.B - coding.B;
        }
        if (n == 0) {
            n = this.H - coding.H;
        }
        if (n == 0) {
            n = this.S - coding.S;
        }
        return n;
    }
    
    public int distanceFrom(Coding coding) {
        int n = this.del - coding.del;
        if (n < 0) {
            n = -n;
        }
        int n2 = this.S - coding.S;
        if (n2 < 0) {
            n2 = -n2;
        }
        int n3 = this.B - coding.B;
        if (n3 < 0) {
            n3 = -n3;
        }
        int n6;
        if (this.H == coding.H) {
            n6 = 0;
        }
        else {
            int hl = this.getHL();
            int hl2 = coding.getHL();
            int n4 = hl * hl;
            int n5 = hl2 * hl2;
            if (n4 > n5) {
                n6 = ceil_lg2(1 + (n4 - 1) / n5);
            }
            else {
                n6 = ceil_lg2(1 + (n5 - 1) / n4);
            }
        }
        int n7 = 5 * (n + n2 + n3) + n6;
        assert this.compareTo(coding) == 0;
        return n7;
    }
    
    private int getHL() {
        if (this.H <= 128) {
            return this.H;
        }
        if (this.L >= 1) {
            return 16384 / this.L;
        }
        return 32768;
    }
    
    static int ceil_lg2(int i) {
        assert i - 1 >= 0;
        --i;
        int n = 0;
        while (i != 0) {
            ++n;
            i >>= 1;
        }
        return n;
    }
    
    static int bitWidth(int n) {
        if (n < 0) {
            n ^= -1;
        }
        int n2 = 0;
        int n3 = n;
        if (n3 < Coding.byteBitWidths.length) {
            return Coding.byteBitWidths[n3];
        }
        int n4 = n3 >>> 16;
        if (n4 != 0) {
            n3 = n4;
            n2 += 16;
        }
        int n5 = n3 >>> 8;
        if (n5 != 0) {
            n3 = n5;
            n2 += 8;
        }
        return n2 + Coding.byteBitWidths[n3];
    }
    
    static int[] makeDeltas(int[] array, int n, int n2, int n3, int n4) {
        assert n4 >= n3;
        int n5 = n2 - n;
        int[] array2 = new int[n5];
        int n6 = 0;
        if (n3 == n4) {
            for (int j = 0; j < n5; ++j) {
                int n9 = array[n + j];
                array2[j] = n9 - n6;
                n6 = n9;
            }
            return array2;
        }
        int i = 0;
        while (i < n5) {
            int n7 = array[n + i];
            assert n7 >= 0 && n7 + n3 <= n4;
            int n8 = n7 - n6;
            assert (long)n8 == (long)n7 - (long)n6;
            n6 = n7;
            array2[i] = reduceToSignedRange(n8, n3, n4);
            ++i;
        }
        return array2;
    }
    
    boolean canRepresent(int n, int n2) {
        assert n <= n2;
        if (this.del > 0) {
            if (this.isSubrange()) {
                return this.canRepresentUnsigned(n2) && this.canRepresentUnsigned(n);
            }
            return this.isFullRange();
        }
        else {
            return this.canRepresentSigned(n2) && this.canRepresentSigned(n);
        }
    }
    
    boolean canRepresent(int[] array, int n, int n2) {
        int n3 = n2 - n;
        if (n3 == 0) {
            return true;
        }
        if (this.isFullRange()) {
            return true;
        }
        int n4 = array[n];
        for (int i = 1; i < n3; ++i) {
            int n5 = array[n + i];
            if (n4 < n5) {
                n4 = n5;
            }
            if (n4 > n5) {
                n4 = n5;
            }
        }
        return this.canRepresent(n4, n4);
    }
    
    @Override
    public double getBitLength(int n) {
        return (double)this.getLength(n) * 8.0;
    }
    
    public int getLength(int reduceToSignedRange) {
        if (this.isDelta() && this.isSubrange()) {
            if (!this.canRepresentUnsigned(reduceToSignedRange)) {
                return Integer.MAX_VALUE;
            }
            reduceToSignedRange = this.reduceToSignedRange(reduceToSignedRange);
        }
        if (reduceToSignedRange >= 0) {
            int j = 0;
            while (j < this.B) {
                if (reduceToSignedRange <= this.byteMax[j]) {
                    return j + 1;
                }
                ++j;
            }
        }
        else {
            int i = 0;
            while (i < this.B) {
                if (reduceToSignedRange >= this.byteMin[i]) {
                    return i + 1;
                }
                ++i;
            }
        }
        return Integer.MAX_VALUE;
    }
    
    public int getLength(int[] array, int n, int n2) {
        int n3 = n2 - n;
        if (this.B == 1) {
            return n3;
        }
        if (this.L == 0) {
            return n3 * this.B;
        }
        if (this.isDelta()) {
            int[] array2;
            if (!this.isSubrange()) {
                array2 = makeDeltas(array, n, n2, 0, 0);
            }
            else {
                array2 = makeDeltas(array, n, n2, this.min, this.max);
            }
            array = array2;
            n = 0;
        }
        int i = 1;
        while (i <= this.B) {
            int n4 = this.byteMax[i - 1];
            int n5 = this.byteMin[i - 1];
            int n6 = 0;
            for (int j = 0; j < n3; ++j) {
                int n7 = array[n + j];
                if (n7 >= 0) {
                    if (n7 > n4) {
                        ++n6;
                    }
                }
                else if (n7 < n5) {
                    ++n6;
                }
            }
            if (n6 == 0) {
                break;
            }
            if (i == this.B) {
                return Integer.MAX_VALUE;
            }
            n3 += n6;
            ++i;
        }
        return n3;
    }
    
    @Override
    public byte[] getMetaCoding(Coding coding) {
        if (coding == this) {
            return new byte[] { 0 };
        }
        int index = BandStructure.indexOf(this);
        if (index > 0) {
            return new byte[] { (byte)index };
        }
        return new byte[] { 116, (byte)(this.del + 2 * this.S + 8 * (this.B - 1)), (byte)(this.H - 1) };
    }
    
    public static int parseMetaCoding(byte[] array, int n, Coding coding, CodingMethod[] array2) {
        int n2 = array[n++] & 0xFF;
        if (1 <= n2 && n2 <= 115) {
            Coding codingForIndex = BandStructure.codingForIndex(n2);
            assert codingForIndex != null;
            array2[0] = codingForIndex;
            return n;
        }
        else if (n2 == 116) {
            int n3 = array[n++] & 0xFF;
            int n4 = array[n++] & 0xFF;
            int i = n3 % 2;
            int j = n3 / 2 % 4;
            int k = n3 / 8 + 1;
            int l = n4 + 1;
            if (1 > k || k > 5 || 0 > j || j > 2 || 1 > l || l > 256 || 0 > i || i > 1 || (k == 1 && l != 256) || (k == 5 && l == 256)) {
                throw new RuntimeException("Bad arb. coding: (" + k + "," + l + "," + j + "," + i);
            }
            array2[0] = of(k, l, j, i);
            return n;
        }
        else {
            return n - 1;
        }
    }
    
    public String keyString() {
        return "(" + this.B + "," + this.H + "," + this.S + "," + this.del + ")";
    }
    
    @Override
    public String toString() {
        return "Coding" + this.keyString();
    }
    
    String stringForDebug() {
        String str = this.keyString() + " L=" + this.L + " r=[" + ((this.min == Integer.MIN_VALUE) ? "min" : ("" + this.min)) + "," + ((this.max == Integer.MAX_VALUE) ? "max" : ("" + this.max)) + "]";
        if (this.isSubrange()) {
            str += " subrange";
        }
        else if (!this.isFullRange()) {
            str += " MIDRANGE";
        }
        if (Coding.verboseStringForDebug) {
            String s = str + " {";
            int n = 0;
            int i = 1;
            while (i <= this.B) {
                int saturate32 = saturate32((long)this.byteMax[i - 1] - (long)this.byteMin[i - 1] + 1L);
                assert saturate32 == saturate32(codeRangeLong(this.B, this.H, i));
                int j = saturate32 - n;
                s = s + " #" + i + "=" + (((n = j) == Integer.MAX_VALUE) ? "max" : ("" + j));
                ++i;
            }
            str = s + " }";
            return str;
        }
        return str;
    }
    
    static {
        byteBitWidths = new byte[256];
        for (int i = 0; i < Coding.byteBitWidths.length; ++i) {
            Coding.byteBitWidths[i] = (byte)ceil_lg2(i + 1);
        }
        int j = 10;
        while (j >= 0) {
            assert bitWidth(j) == ceil_lg2(j + 1);
            j = (j << 1) - (j >> 3);
        }
        Coding.verboseStringForDebug = false;
    }
}
